{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeOperators #-}

{-# OPTIONS_GHC -fno-warn-orphans #-}

{-|

  Copyright: 2017 (C) AlphaSheets, Inc
  Description: Client support for multipart.

-}

module BBW.Multipart where

import Control.Monad
import Control.Monad.IO.Class (MonadIO(..))
import Data.Monoid
import Data.Sequence as Seq
import Network.HTTP.Client.MultipartFormData (webkitBoundary)
import Network.HTTP.Media.MediaType
import Network.HTTP.Types
import Servant.API
import Servant.Client.Core.Internal.HasClient
import Servant.Client.Core.Internal.Request
import Servant.Multipart
import System.FilePath
import Data.Proxy

import qualified Data.ByteString.Lazy as LBS
import qualified Data.CaseInsensitive as CI
import qualified Data.Foldable
import qualified Data.Text as T
import qualified Data.Text.Encoding as TE
import qualified Network.HTTP.Types as HTTP
import qualified Network.Mime as Mime

-- | A single part of a multipart message.
-- Adapted from @Network.HTTP.Client.MultipartFormData.Part@
data Part = Part
    { partName :: T.Text -- ^ Name of the corresponding \<input\>
    , partFilename :: Maybe String -- ^ A file name, if this is an attached file
    , partContentType :: Maybe Mime.MimeType -- ^ Content type
    , partHeaders :: [HTTP.Header] -- ^ List of additional headers
    , partGetBody :: IO LBS.ByteString -- ^ Action in m which returns the body
                                   -- of a message.
                                   -- XXX: This is servant' requestbody
    }

-- | A class for encoding datatypes as 'Part's.
--  See 'Part', 'partInput', 'partFile'
class ToMultipart a where
  toMultipart :: a -> [Part]

-- TODO: _inputs are ignored
-- TODO: Mem could be generalized using MultipartBackend
instance ToMultipart (MultipartData Mem) where
  toMultipart (MultipartData _inputs files) =
    map fileDataToPart files
    where
      fileDataToPart :: FileData Mem -> Part
      fileDataToPart fd = Part
        { partName = fdInputName fd
        , partFilename = Just (T.unpack (fdFileName fd))
        , partContentType = Just (TE.encodeUtf8 (fdFileCType fd))
        , partHeaders = []
        , partGetBody = pure (fdPayload fd)
        }

-- A somewhat meaningful instance of 'HasClient' for multipart data. Note that
-- when a file 'Part' is encountered, the whole file will be loaded up in
-- memory. As such,
--
--  !!!!THIS SHOULD BE USED WITH CAUTION!!!!
--
-- See https://github.com/haskell-servant/servant/issues/886
--
instance (ToMultipart ty, Client m sub ~ m a, MonadIO m, HasClient m sub) =>
    HasClient m
      (MultipartForm Tmp ty :> sub) where
  type
    Client m
      (MultipartForm Tmp ty :> sub)
      = ty -> Client m sub
  clientWithRoute
    :: Proxy m
    -> Proxy (MultipartForm Tmp ty :> sub)
    -> Request
    -> ty -> Client m sub
  clientWithRoute pMonad _p2 req dfi = do
      -- Adapted from @Network.HTTP.Client.MultipartFormData.formDataBody@
      boundary <- liftIO webkitBoundary
      body <- liftIO $ renderParts (LBS.fromStrict boundary) $ toMultipart dfi
      let req' = req
            { requestHeaders =
                Seq.filter (\(x, _) -> x /= hContentType) (requestHeaders req)
            , requestBody = Just
                ( RequestBodyLBS body
                , "multipart" // "form-data" /: ("boundary", boundary)
                )
            }
      clientWithRoute pMonad (Proxy :: Proxy sub) req'


-------------------------------------------------------------------------------
-- XXX: The functions 'renderParts' and 'renderPart' were adapted from
-- @Network.HTTP.Client.MultipartFormData@.
-------------------------------------------------------------------------------

-- | Combine the 'Part's to form multipart/form-data body
renderParts :: LBS.ByteString    -- ^ Boundary between parts.
            -> [Part] -> IO LBS.ByteString
renderParts boundary parts = (fin . mconcat) `liftM` mapM (renderPart boundary) parts
  where fin = (<> "--" <> boundary <> "--\r\n")

renderPart :: LBS.ByteString     -- ^ Boundary between parts.
           -> Part -> IO LBS.ByteString
renderPart boundary (Part name mfilename mcontenttype hdrs get) =
    fmap render get
  where render renderBody =
            "--" <> boundary <> "\r\n"
         <> "Content-Disposition: form-data; name=\""
         <> LBS.fromStrict (TE.encodeUtf8 name)
         <> (case mfilename of
                 Just f -> "\"; filename=\""
                        <> LBS.fromStrict (TE.encodeUtf8 (T.pack $ takeFileName f))
                 _ -> mempty)
         <> "\""
         <> (case mcontenttype of
                Just ct -> "\r\n"
                        <> "Content-Type: "
                        <> LBS.fromStrict ct
                _ -> mempty)
         <> Data.Foldable.foldMap (\(k, v) ->
               "\r\n"
            <> LBS.fromStrict (CI.original k)
            <> ": "
            <> LBS.fromStrict v) hdrs
         <> "\r\n\r\n"
         <> renderBody <> "\r\n"

-- | Create a key-value 'Part'.
partInput :: T.Text -> LBS.ByteString -> Part
partInput n b = Part
  { partName = n
  , partFilename = Nothing
  , partContentType = Nothing
  , partHeaders = []
  , partGetBody = pure b
  }

-- | Create a 'Part' for a local file. This does not infer the mime type of the
-- file, although one may be provided.
partFile
  :: T.Text
    -- ^ The part name
  -> FilePath
    -- ^ The path to the local file
  -> Maybe Mime.MimeType
    -- ^ Optional mime type
  -> Part
partFile n fp mmt = Part
  { partName = n
  , partFilename = Just fp
  , partContentType = mmt
  , partHeaders = []
  , partGetBody = LBS.readFile fp
  }
